(* Time-stamp: <modified the 28/07/2022 (at 10:49) by Erwan Jahier> *)

(* Author: Antony Zahran *)

open Printf
open Types
open Parser

let yaml_file = ref ""


(* let print_var_outputs name var  = 
*   fprintf cfile "void %s_TASK_getout_%s(%s_TASK_type tD, %s* p%s) {\n" name var.name name var.var_type var.name;
*   fprintf cfile "%s_TASK_struct* ts = (%s_TASK_struct*) tD;\n" name;
*   fprintf cfile "*p%s = ts->ctx.%s;\n" var.name var.name;
*   fprintf cfile "}\n\n";
* in *)
let main () =
  if (Array.length Sys.argv) <= 1 then (
    Arg.usage MainArgs.speclist MainArgs.usage; flush stdout; exit 2
  ) else (
    try Arg.parse MainArgs.speclist (fun s -> yaml_file := s) MainArgs.usage
    with  
    | Failure(e) -> print_string e; flush stdout; flush stderr; exit 2
    | e -> print_string (Printexc.to_string e);  flush stdout; exit 2
  );
  let yaml = Yaml_unix.of_file_exn Fpath.(v !yaml_file) in
  Parser.get_data yaml;


  (* creates the .c file *)
  let cfile = open_out (!Parser.name ^ "_thread.c") in

  (* includes *)
  fprintf cfile "#include <stdio.h>\n";
  fprintf cfile "#include <pthread.h>\n";
  fprintf cfile "#include <semaphore.h>\n";
  fprintf cfile "#include <errno.h>\n";
  fprintf cfile "#include <time.h>\n";
  fprintf cfile "#include <sys/time.h>\n";

  fprintf cfile "#include \"%s.h\"\n" !Parser.name;
  fprintf cfile "\n";

  (* semaphores macro *)
  fprintf cfile "/* Initialize the semaphore with the given count.  */\n";
  fprintf cfile "#define SEM_INIT(sem, v, max_v) sem_init(&(sem), 0, (v))\n";
  fprintf cfile "/* wait for the semaphore to be active (i.e. equal to 1) and then decrement it by one. */\n";
  fprintf cfile "#define SEM_WAIT(sem) while (sem_wait(&sem) != 0 && errno == EINTR) continue\n";
  fprintf cfile "/* make the semaphore active (i.e. equal to 1) */\n";
  fprintf cfile "#define SEM_SIGNAL(sem) sem_post(&(sem))\n";
  fprintf cfile "\n";



  fprintf cfile "/* task structures */\n";
  let print_task_structs task =
    fprintf cfile "typedef struct {\n";
    fprintf cfile "\t%s_ctx_type ctx;\n" task.name;
    fprintf cfile "\tsem_t sem_start;\n";
    fprintf cfile "\tsem_t sem_join;\n";
    fprintf cfile "pthread_t thr;\n}";
    fprintf cfile "%s_TASK_struct;" task.name;
    fprintf cfile "\n\n"
  in
  List.iter print_task_structs !tasks;

  fprintf cfile "/* runners */\n";
  let print_task_runners task = 
    fprintf cfile "void* %s_runner(void* cd) {\n" task.name;
    fprintf cfile "\t%s_TASK_struct* ts = (%s_TASK_struct*) cd;\n" task.name task.name;
    fprintf cfile "\twhile(1){\n";
    fprintf cfile "\t\tSEM_WAIT(ts->sem_start);\n";
    fprintf cfile "\t\t%s_step(&ts->ctx);\n" task.name;
    fprintf cfile "\t\tSEM_SIGNAL(ts->sem_join);\n";
    fprintf cfile "\t}\n";
    fprintf cfile "}\n\n"
  in
  List.iter print_task_runners !tasks;

  fprintf cfile "/* task Initializers */\n";
  let print_task_inits task = 
    fprintf cfile "%s_TASK_type %s_TASK_init() {\n" task.name task.name;
    fprintf cfile "\t%s_TASK_struct* ts = (%s_TASK_struct*)calloc(1, sizeof(%s_TASK_struct));\n" task.name task.name task.name;
    fprintf cfile "\t%s_ctx_init(&ts->ctx);\n" task.name;
    fprintf cfile "\tSEM_INIT(ts->sem_start, 0, 1);\n";
    fprintf cfile "\tSEM_INIT(ts->sem_join, 0, 1);\n";
    fprintf cfile "\tpthread_create(&(ts->thr), NULL, %s_runner, ts);\n" task.name;
    fprintf cfile "\treturn (void*) ts;\n";
    fprintf cfile "}\n\n"
  in
  List.iter print_task_inits !tasks;
  
  fprintf cfile "/* task resets */\n";
  let print_task_resets task =
    fprintf cfile "void %s_TASK_reset(%s_TASK_type tD) {\n" task.name task.name;
    fprintf cfile "\t%s_TASK_struct* ts = (%s_TASK_struct*) tD;\n" task.name task.name;
    fprintf cfile "\t%s_ctx_reset(&ts->ctx);\n" task.name;
    fprintf cfile "}\n\n"
  in
  List.iter print_task_resets !tasks; 

  fprintf cfile "/* task start */\n";
  let print_task_starts task =
    fprintf cfile "void %s_TASK_START(%s_TASK_type tD) {\n" task.name task.name;
    fprintf cfile "\t%s_TASK_struct* ts = (%s_TASK_struct*) tD;\n" task.name task.name;
    fprintf cfile "\tSEM_SIGNAL(ts->sem_start);\n";
    fprintf cfile "}\n\n"
  in
  List.iter print_task_starts !tasks; 

  fprintf cfile "/* task join */\n";
  let print_task_joins task =
    fprintf cfile "void %s_TASK_JOIN(%s_TASK_type tD) {\n" task.name task.name;
    fprintf cfile "\t%s_TASK_struct* ts = (%s_TASK_struct*) tD;\n" task.name task.name;
    fprintf cfile "\tSEM_WAIT(ts->sem_join);\n";
    fprintf cfile "}\n\n"
  in
  List.iter print_task_joins !tasks; 
  
  fprintf cfile "/* task inputs */\n";
  let print_task_inputs task = 
    let (print_var_inputs : string -> variable -> unit) = 
      fun name var -> 
      fprintf cfile "void %s_TASK_setin_%s(%s_TASK_type tD, %s %s) {\n" name var.name name var.var_type var.name;
      fprintf cfile "\t%s_TASK_struct* ts = (%s_TASK_struct*) tD;\n" name name;
      fprintf cfile "\tts->ctx.%s = %s;\n" var.name var.name;
      fprintf cfile "}\n\n";
    in
    List.iter (print_var_inputs task.name) task.var_in;
  in
  List.iter print_task_inputs !tasks;

  fprintf cfile "/* task outputs */\n";
  let print_task_outputs task = 
    let (print_var_outputs : string -> variable -> unit) = 
      fun name var -> 
      fprintf cfile "void %s_TASK_getout_%s(%s_TASK_type tD, %s* p%s) {\n" name var.name name var.var_type var.name;
      fprintf cfile "\t%s_TASK_struct* ts = (%s_TASK_struct*) tD;\n" name name;
      fprintf cfile "\t*p%s = ts->ctx.%s;\n" var.name var.name;
      fprintf cfile "}\n\n";
    in
    List.iter (print_var_outputs task.name) task.var_out;
  in
  List.iter print_task_outputs !tasks;





(*
*  (* main function *)
*  fprintf cfile "/* Main function */\n";
*  fprintf cfile "void main() {
*  int _s = 0;\n";
*
*  fprintf cfile "    /* Initialize context */\n";
*  fprintf cfile "    ctx = %s(NULL);\n" main.ctx_new;
*  fprintf cfile "\n";
*
*  fprintf cfile "    /* Initialize semaphores */\n";
*  List.iter (fprintf cfile "    SEM_INIT(channel%i, 0, 1);\n") !channels;
*  fprintf cfile "    \n";
*
*  let x = List.length !instances in
*  let i = x+1 in
*  fprintf cfile "\t#pragma omp parallel num_threads(%i) default(shared)\n" i;
*  fprintf cfile "\t{\n";
*  (* main loop *)
*  fprintf cfile "\t\t#pragma omp single nowait\n{\n";
*  fprintf cfile "print_rif_declaration();\n";
*  output_string cfile "    /* Main loop */
* while(true) {
*        if (ISATTY) printf(\"#step \\%d \\n\", _s+1);
*        else if(_s) printf(\"\\n\");
*        fflush(stdout);
*        ++_s;\n";
*  fprintf cfile "        get_inputs(ctx";
*  List.iter (fun (x:variable) -> fprintf cfile ", &%s" x.name) main.var_in;
*  fprintf cfile ");\n";
*
*  List.iter (fprintf cfile "        SEM_SIGNAL(channel%i);\n") main.ch_in;
*  List.iter (fprintf cfile "        SEM_WAIT(channel%i);\n") main.ch_out;
*
*  fprintf cfile "        print_outputs(%s" (List.hd main.var_out).name;
*  List.iter (fun (x:variable) -> fprintf cfile ", %s" x.name) (List.tl main.var_out);
*  fprintf cfile ");\n";
*
*  fprintf cfile "    }\n";
*  fprintf cfile "\t\t}\n";
*  fprintf cfile "    \n";
*  
*  let print_instance_thread instance = 
*    fprintf cfile "\n#pragma omp single nowait\n{\n";
*    fprintf cfile "loop_%s%i();" instance.node.file_name instance.id;
*    fprintf cfile "}\n";
*  in
*  List.iter print_instance_thread !instances;
*
*  fprintf cfile "\t}\n";
*  fprintf cfile "\n"
*  fprintf cfile "}\n";
*)
;;

let _ = main ()
